;;; Test harness based on Allegro's.
;;; $Id: acl-tester.lisp 20 2016-12-04 11:10:54Z digo $
;;
;; copyright (c) 1985-1986 Franz Inc, Alameda, CA
;; copyright (c) 1986-2001 Franz Inc, Berkeley, CA - All rights reserved.
;; copyright (c) 2001-2003 Kevin Rosenberg (portability changes)
;; copyright (c) 2004-2012 Dmitriy Ivanov (small improvements)
;;
;; This code is free software; you can redistribute it and/or
;; modify it under the terms of the version 2.1 of
;; the GNU Lesser General Public License as published by 
;; the Free Software Foundation, as clarified by the Franz
;; preamble to the LGPL found in
;; http://opensource.franz.com/preamble.html.
;;
;; This code is distributed in the hope that it will be useful,
;; but without any warranty; without even the implied warranty of
;; merchantability or fitness for a particular purpose.  See the GNU
;; Lesser General Public License for more details.
;;
;; Version 2.1 of the GNU Lesser General Public License can be
;; found at http://opensource.franz.com/license.html.
;; If it is not present, you can access it from
;; http://www.gnu.org/copyleft/lesser.txt (until superseded by a newer
;; version) or write to the Free Software Foundation, Inc., 59 Temple
;; Place, Suite 330, Boston, MA  02111-1307  USA
;;
;;;; from the original ACL 6.1 sources:
;; Id: tester.cl,v 2.2.12.1 2001/06/05 18:45:10 layer Exp
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; DI Small changes/improvements
;;;  1. The full package name changed from :acl-compat.util.test to
;;;     :acl-compat.util.tester.
;;;  2. The default test predicate changed from #'eql to #'equal.
;;;  3. Slightly altered format control strings for reporting.
;;;  4. Changed with-test to return multiple-values.
;;;  5. Added test-true macro.
;;;  6. Added :ignore value of the multiple-values argument to test results by
;;;     comparing only the first of the values returned with the expected value
;;;     (as if :multiple-values = nil) and ingoring the rest of them
;;;     without indicating the whole test as failed.
;;;  7. Added test-values-ignore macro and :warning value of the (evaluated)
;;;     errorset argument to test.
;;;  8. Added reported-form parameter to the test-error and test-no-error macros.
;;;  9. Rewritten without the errorset macro (if* has been already eliminated by Kevin).
;;; 10. Customized indent for LispWorks editor.

(defpackage :acl-compat.util.tester
  (:use :common-lisp)		; DI 2015-Oct-23
  (:nicknames :util.test :test)
  (:shadow #:test)
  (:export
;;;; Control variables:
   #:*break-on-test-failures*
   #:*error-protect-tests*
   #:*test-errors*
   #:*test-successes*
   #:*test-unexpected-failures*

;;;; The test macros:
   #:test
   #:test-true			; DI 2006-Aug-31
   #:test-error
   #:test-no-error
   #:test-warning
   #:test-no-warning
   
   #:with-tests))

;;; DI 2006-Sep-07: Make standard indent in LispWorks editor equal to 2
#+lispworks (editor:setup-indent "test" 1 2 6)		; indent the first arg by 6
#+lispworks (editor:setup-indent "test-true"       0 2)
#+lispworks (editor:setup-indent "test-error"      0 2)
#+lispworks (editor:setup-indent "test-no-error"   0 2)
#+lispworks (editor:setup-indent "test-warning"    0 2)
#+lispworks (editor:setup-indent "test-no-warning" 0 2)


(in-package :acl-compat.util.tester)

;;; DI 2009-Mar-06: This must be set before defining any test as used for macroexpansion
(defparameter *error-protect-tests* nil
  "Protect each test from errors.  If an error occurs, then that will be
taken as a test failure unless test-error is being used.")

(defvar *break-on-test-failures* nil
  "When a test failure occurs, common-lisp:break is called, allowing
interactive debugging of the failure.")

(defvar *test-errors* 0
  "The value is the number of test errors which have occurred.")
(defvar *test-successes* 0
  "The value is the number of test successes which have occurred.")
(defvar *test-unexpected-failures* 0
  "The value is the number of unexpected test failures which have occurred.")

#-lispworks
(define-condition simple-break (error simple-condition) ())

(defmacro test-values-errorset (form &optional announce catch-breaks)
  ;; internal macro
  (let ((g-announce (gensym))
	(g-catch-breaks (gensym)))
    `(let* ((,g-announce ,announce)
	    (,g-catch-breaks ,catch-breaks))
       (handler-case (cons t (multiple-value-list ,form))
	 (condition (condition)
	   (cond ((and (null ,g-catch-breaks)
                       (typep condition #+lispworks 'conditions:debug-break
                                        #-lispworks 'simple-break))
                  (break condition))
                 (,g-announce
                  (format *error-output* "~&Condition type: ~a~%Message: ~a~%"
                          (class-of condition) condition)))
           condition)))))

(defmacro test-values (form &optional announce catch-breaks)
  ;; internal macro
  (if *error-protect-tests*
      `(test-values-errorset ,form ,announce ,catch-breaks)
      `(cons t (multiple-value-list ,form))))

(defmacro test-values-ignore (form &optional announce catch-breaks muffle-warning)
 ;;; DI 2006-Sep-26: Announce a condition of any type but suppress warnings.
  (let ((g-announce (gensym))
	(g-catch-breaks (gensym))
        (g-muffle-warning (gensym))
        (g-block (gensym)))
    `(block ,g-block
       (let* ((,g-announce ,announce)
              (,g-catch-breaks ,catch-breaks)
              (,g-muffle-warning ,muffle-warning))
         (handler-bind
             ((condition
               (lambda (condition)
                 (if (and (null ,g-catch-breaks)
                          (typep condition #+lispworks 'conditions:debug-break
                                           #-lispworks 'simple-break))
                     (progn (break condition)
                       (return-from ,g-block condition))
                     (progn
                       (when ,g-announce
                         (format *error-output* "~&Condition type: ~a~%Message: ~a~%"
                                 (class-of condition) condition))
                       (if (and ,g-muffle-warning (typep condition 'warning))
                           (muffle-warning condition)
                           (return-from ,g-block condition)))))))
           (cons t (multiple-value-list ,form)))))))

(defmacro test (expected-value test-form
		&key (test #'equal test-given)		; DI 2006-Jun-23: Was #'eql
		     (multiple-values nil multiple-values-given)
		     (fail-info nil fail-info-given)
		     (known-failure nil known-failure-given)

;;;;;;;;;; internal, undocumented keywords:
;;;; Note about these keywords: if they were documented, we'd have a
;;;; problem, since they break the left-to-right order of evaluation.
;;;; Specifically, errorset breaks it, and I don't see any way around
;;;; that. `errorset' is used by the old test.cl module (eg, test-equal-errorset).
;;;; DI 2006-Sep-26:
;;;;   If `errorset' is specified, the test is protected from errors; moreover,
;;;;   if it evaluates to :warning, warnings are announced but muffled,
;;;;   and the test go on executing.
		     errorset
		     reported-form
		     (wanted-message nil wanted-message-given)
		     (got-message nil got-message-given))
  "Perform a single test.  `expected-value' is the reference value for the
test.  `test-form' is a form that will produce the value to be compared to
the expected-value.  If the values are not the same, then an error is
logged, otherwise a success is logged.

Normally the comparison of values is done with `equal'.  The `test' keyword
argument can be used to specify other comparison functions, such as eq,
eql, equalp, string=, string-equal, etc.

Normally, only the first return value from the test-form is considered,
however if `multiple-values' is t, then all values returned from test-form
are considered. If `multiple-values' is :ignore, the extra values are skipped.

`fail-info' allows more information to be printed with a test failure.

`known-failure' marks the test as a known failure.  This allows for
programs that do regression analysis on the output from a test run to
discriminate on new versus known failures."
  `(test-check
    :expected-result ,expected-value
    :test-results
    ,(if errorset
         `(test-values-ignore ,test-form t nil (eq ,errorset :warning))
         `(test-values ,test-form t))
    ,@(when test-given `(:predicate ,test))
    ,@(when multiple-values-given `(:multiple-values ,multiple-values))
    ,@(when fail-info-given `(:fail-info ,fail-info))
    ,@(when known-failure-given `(:known-failure ,known-failure))
    :test-form ',(if reported-form reported-form test-form)
    ,@(when wanted-message-given `(:wanted-message ,wanted-message))
    ,@(when got-message-given `(:got-message ,got-message))))

;;; DI 2006-Aug-31
(defun iff (arg1 arg2) (if arg1 arg2 (not arg2)))
(defmacro test-true (test-form &rest args) `(test ,t ,test-form ,:test #'iff ,@args))

(defgeneric conditionp (thing)
 (:method ((thing condition)) t)
 (:method ((thing t)) nil))

(defmacro test-error (form &key announce
				catch-breaks
				(fail-info nil fail-info-given)
				(known-failure nil known-failure-given)
                                reported-form			; DI 2006-Sep-04
				(condition-type ''simple-error)
				(include-subtypes nil include-subtypes-given)
				(format-control nil format-control-given)
				(format-arguments nil format-arguments-given))
  "Test that `form' signals an error. The order of evaluation of the
arguments is keywords first, then test form.

If `announce' is non-nil, then cause the error message to be printed.

The `catch-breaks' is non-nil then consider a call to common-lisp:break an
`error'.

`fail-info' allows more information to be printed with a test failure.

`known-failure' marks the test as a known failure.  This allows for
programs that do regression analysis on the output from a test run to
discriminate on new versus known failures.

If `condition-type' is non-nil, it should be a symbol naming a condition
type, which is used to check against the signalled condition type.  The
test will fail if they do not match.

If `include-subtypes' is true, allow the caught condition to be a subclass
of the `condition-type' given.

`format-control' and `format-arguments' can be used to check the error
message itself."
  (let ((g-announce (gensym))
	(g-catch-breaks (gensym))
	(g-fail-info (gensym))
	(g-known-failure (gensym))
	(g-condition-type (gensym))
	(g-include-subtypes (gensym))
	(g-format-control (gensym))
	(g-format-arguments (gensym))
	(g-c (gensym)))
    `(let* ((,g-announce ,announce)
	    (,g-catch-breaks ,catch-breaks)
	    ,@(when fail-info-given `((,g-fail-info ,fail-info)))
	    ,@(when known-failure-given `((,g-known-failure ,known-failure)))
	    (,g-condition-type ,condition-type)
	    ,@(when include-subtypes-given
		`((,g-include-subtypes ,include-subtypes)))
	    ,@(when format-control-given
		`((,g-format-control ,format-control)))
	    ,@(when format-arguments-given
		`((,g-format-arguments ,format-arguments)))
	    (,g-c (test-values-errorset ,form ,g-announce ,g-catch-breaks)))
       (test-check
	:predicate #'eq
	:expected-result t
	:test-results
	(test-values (and (conditionp ,g-c)
			  ,@(if include-subtypes-given
			        `((if ,g-include-subtypes
                                      (typep ,g-c ,g-condition-type)
                                      (eq (type-of ,g-c) ,g-condition-type)))
                                `((eq (type-of ,g-c) ,g-condition-type)))
			  ,@(when format-control-given
			      `((or (null ,g-format-control)
                                    (string=
                                     (concatenate 'simple-string
                                                  "~1@<" ,g-format-control "~:@>")
                                     (simple-condition-format-control ,g-c)))))
			  ,@(when format-arguments-given
			      `((or (null ,g-format-arguments)
                                    (equal ,g-format-arguments
                                           (simple-condition-format-arguments ,g-c))))))
		     t)
	:test-form ',(or reported-form form)			; DI 2006-Sep-04
	,@(when fail-info-given `(:fail-info ,g-fail-info))
	,@(when known-failure-given `(:known-failure ,g-known-failure))
	:condition-type ,g-condition-type
	:condition ,g-c
	,@(when include-subtypes-given
	    `(:include-subtypes ,g-include-subtypes))
	,@(when format-control-given
	    `(:format-control ,g-format-control))
	,@(when format-arguments-given
	    `(:format-arguments ,g-format-arguments))))))

(defmacro test-no-error (form &key announce
				   catch-breaks
				   (fail-info nil fail-info-given)
				   (known-failure nil known-failure-given)
                                   reported-form)		; DI 2006-Sep-04
  "Test that `form' does not signal an error.  The order of evaluation of
the arguments is keywords first, then test form.

If `announce' is non-nil, then cause the error message to be printed.

The `catch-breaks' is non-nil then consider a call to common-lisp:break an
`error'.

`fail-info' allows more information to be printed with a test failure.

`known-failure' marks the test as a known failure.  This allows for
programs that do regression analysis on the output from a test run to
discriminate on new versus known failures."
  (let ((g-announce (gensym))
	(g-catch-breaks (gensym))
	(g-fail-info (gensym))
	(g-known-failure (gensym))
	(g-c (gensym)))
    `(let* ((,g-announce ,announce)
	    (,g-catch-breaks ,catch-breaks)
	    ,@(when fail-info-given `((,g-fail-info ,fail-info)))
	    ,@(when known-failure-given `((,g-known-failure ,known-failure)))
	    (,g-c (test-values-errorset ,form ,g-announce ,g-catch-breaks)))
       (test-check
	:predicate #'eq
	:expected-result t
	:test-results (test-values (not (conditionp ,g-c)))
	:condition ,g-c
	:test-form ',(or reported-form form)			; DI 2006-Sep-04
	,@(when fail-info-given `(:fail-info ,g-fail-info))
	,@(when known-failure-given `(:known-failure ,g-known-failure))))))

(defvar *warn-cookie* (cons nil nil))

(defmacro test-warning (form &key fail-info known-failure)
  "Test that `form' signals a warning.  The order of evaluation of
the arguments is keywords first, then test form.

`fail-info' allows more information to be printed with a test failure.

`known-failure' marks the test as a known failure.  This allows for
programs that do regression analysis on the output from a test run to
discriminate on new versus known failures."
  (let ((g-fail-info (gensym))
	(g-known-failure (gensym))
	(g-value (gensym)))
    `(let* ((,g-fail-info ,fail-info)
	    (,g-known-failure ,known-failure)
	    (,g-value (test-values-errorset ,form nil t)))
       (test
	*warn-cookie*
	(if (or (typep ,g-value 'simple-warning) (typep ,g-value 'warning))
	    *warn-cookie*
	    nil)			; test produced no warning
	:test #'eq
	:reported-form ,form		; quoted by test macro
	:wanted-message "a warning"
	:got-message "no warning"
	:fail-info ,g-fail-info
	:known-failure ,g-known-failure))))

(defmacro test-no-warning (form &key fail-info known-failure)
  "Test that `form' does not signal a warning.  The order of evaluation of
the arguments is keywords first, then test form.

`fail-info' allows more information to be printed with a test failure.

`known-failure' marks the test as a known failure.  This allows for
programs that do regression analysis on the output from a test run to
discriminate on new versus known failures."
  (let ((g-fail-info (gensym))
	(g-known-failure (gensym))
	(g-value (gensym)))
    `(let* ((,g-fail-info ,fail-info)
	    (,g-known-failure ,known-failure)
	    (,g-value (test-values-errorset ,form nil t)))
       (test
	*warn-cookie*
	(if (or (typep ,g-value 'simple-warning) (typep ,g-value 'warning))
	    nil				; test produced warning
	    *warn-cookie*)
	:test #'eq
	:reported-form ',form
	:wanted-message "no warning"
	:got-message "a warning"
	:fail-info ,g-fail-info
	:known-failure ,g-known-failure))))

(defvar *announce-test* nil) ;; if true announce each test that was done

(defun test-check (&key (predicate #'equal)			; DI 2006-Jun-23: Was #'eql
			expected-result test-results test-form
			multiple-values fail-info known-failure
			wanted-message got-message condition-type condition
			include-subtypes format-control format-arguments
		   &aux fail predicate-failed got wanted)
  ;; DI 2012-Nov-30: The call of errorset has replaced with ignore-errors
  ;; for debugging large/complex test sets:
  (when *announce-test*
    (format t "~&Just did test ~s" test-form)
    (force-output))
  
  ;; this is an internal function
  (flet ((check (expected-result result)
	   ;(let* ((results (multiple-value-list
           ;                 (errorset (funcall predicate expected-result result) t)))
	   ;	   (failed (null (car results))))
           (multiple-value-bind (value condition)
               (funcall predicate expected-result result)
	     (if (typep condition 'condition)
		 (progn (format *error-output*			; as per announce t
                         "~&*** Predicate ~s has signaled the error of type ~a
Message: ~a~%Test form: ~s~%"
                         predicate (class-of condition) condition test-form)
                   (setq predicate-failed t)
                   nil)
		 value))))					;(cadr results)
    (when (conditionp test-results)
      (setq condition test-results)
      (setq test-results nil))
    (when (null (car test-results))
      (setq fail t))
    (if (and (not fail) (or (not multiple-values)
                            (eq multiple-values :ignore)))	; DI 2006-Sep-01
        ;; should be a single result
        ;; expected-result is the single result wanted
        (progn
          (when (not (and (cdr test-results)
                          (check expected-result (cadr test-results))))
            (setq fail t))
          (when (and (not fail) (cddr test-results)
                     (not (eq multiple-values :ignore)))	; DI 2006-Sep-01
            (setq fail 'single-got-multiple)))
         ;; multiple results wanted
         ;; expected-result is a list of results, each of which
         ;; should be checked against the corresponding test-results
         ;; using the predicate
         (do ((got (cdr test-results) (cdr got))
              (want expected-result (cdr want)))
             ((or (null got) (null want))
              (when (not (and (null want) (null got)))
                (setq fail t)))
           (when (not (check (car got) (car want)))
             (return (setq fail t)))))
    (when (eq multiple-values :ignore)				; DI 2006-Sep-01
      (setq multiple-values nil))				; act as for a single value
    (cond (fail
           (when (not known-failure)
	      (format *error-output*
	      "~&				    *** UNEXPECTED TEST FAILURE ***~%")
	      (incf *test-unexpected-failures*))
           (format *error-output* "~&Test failed: ~@[known failure: ~*~]~s~%"
                   known-failure test-form)
           (cond ((eq 'single-got-multiple fail)
                  (format *error-output*
                          "~&Reason: additional value were returned from test form.~%"))
                 (predicate-failed
                  (format *error-output* "Reason: predicate error.~%"))
                 ((null (car test-results))
                  (format *error-output*
                          "~&Reason: an error~@[ (of type `~s')~] was detected.~%"
                          (when condition (class-of condition))))
                 (condition
                  (cond ((not (conditionp condition))
                         (format *error-output*
                          "~&Reason: expected but did not detect an error of type `~s'.~%"
				    condition-type))
                        ((null condition-type)
                         (format *error-output*
                                 "~&Reason: detected an unexpected error of type `~s':
        ~a.~%"
                                 (class-of condition)
                                 condition))
                        ((not (if include-subtypes
				  (typep condition condition-type)
				  (eq (class-of condition) (find-class condition-type))))
                         (format *error-output*
                                 "~&Reason: detected an incorrect condition type.
  wanted: ~s~%     got: ~s~%"
                                 condition-type (class-of condition)))
                        ((and format-control
                              (not (string= (setq got
                                                  (concatenate 'simple-string
                                                           "~1@<" format-control "~:@>"))
                                            (setq wanted (simple-condition-format-control
                                                          condition)))))
                         ;; format control doesn't match
                         (format *error-output*
                                 "~&Reason: the format-control was incorrect.
  wanted: ~s~%     got: ~s~%"
                                 wanted got))
                        ((and format-arguments
                              (not (equal (setq got format-arguments)
                                          (setq wanted (simple-condition-format-arguments
                                                        condition)))))
                         (format *error-output* 
                                 "~&Reason: the format-arguments were incorrect.
  wanted: ~s~%     got: ~s~%"
                                 wanted got))
                        (t ;; what else????
                           (error "internal-error"))))
                 (t
                  (let ((*print-length* 50)
                        (*print-level* 10))
                    (cond (wanted-message
                           (format *error-output* "  wanted: ~a~%" wanted-message))
                          ((not multiple-values)
                           (format *error-output* "  wanted: ~s~%" expected-result))
                          (t
                           (format *error-output* "  wanted values: ~{~s~^, ~}~%"
                                   expected-result)))
                    (cond (got-message
                           (format *error-output* "     got: ~a~%" got-message))
                          ((not multiple-values)
                           (format *error-output* "     got: ~s~%" (second test-results)))
                          (t
                           (format *error-output* "     got values: ~{~s~^, ~}~%"
                                   (cdr test-results)))))))
           (when fail-info
             (format *error-output* "Additional info: ~a~%" fail-info))
           (incf *test-errors*)
           (when *break-on-test-failures*
             (break "~a is non-nil." '*break-on-test-failures*)))
          (t
           (cond (known-failure
                  (format *error-output*
                          "~&Expected test failure for ~s did not occur.~%"
                          test-form)
                  (when fail-info
                    (format *error-output* "Additional info: ~a~%" fail-info))
                  (setq fail t))
                 (*announce-test*			; DI 2006-Aug-31
                  (write-string " - success.")))
           (incf *test-successes*)))
    (not fail)))

(defmacro with-tests ((&key (name "unnamed") values) &body body)
 ;;; DI 2006-Jun-20: Return multiple values:
  ;;  1>     The primary value of the last form in the BODY;
  ;;  2,3,4> *test-errors*, *test-unexpected-failures*, and *test-successes*;
  ;;  ...>   More produced by forms passed via the VALUES parameter.
  (let ((g-name (gensym))
        (g-value (gensym)))
    `(let ((,g-name ,name)
           (*test-errors* 0)
           (*test-successes* 0)
           (*test-unexpected-failures* 0))
       (format *error-output* "~&**** Begin test ~a~%" ,g-name)
       (flet ((doit () ,@body))
         (let* ((,g-value
                 (if *break-on-test-failures*
                     (doit)
                     (handler-case (doit)
                       (error (c) (format *error-output*
                         "~&**** Test ~a aborted by signalling an uncaught error:~%~a~%"
                                         ,g-name c)))))
                (state (gsgc-switch :print)))
           (setf (gsgc-switch :print) nil)
           (format *error-output*
                   "~&**** End test ~a
**** Errors detected in this test: ~s~[~:;, UNEXPECTED: ~:*~s~]
**** Successes this test: ~s~%"
                   ,g-name *test-errors* *test-unexpected-failures* *test-successes*)
           (setf (gsgc-switch :print) state)
           (values ,g-value
                   *test-errors* *test-unexpected-failures* *test-successes*
                   ,@values))))))

(provide :tester #+module-versions 1.1)

(defun gsgc-switch (switch)
  #-(or cmu allegro) (declare (ignore switch))
  #+cmu ext:*gc-verbose*
  #+allegro (sys:gsgc-switch switch))

(defun (setf gsgc-switch) (state switch)
  #-(or cmu allegro) (declare (ignore state switch))
  #+cmu (setf ext:*gc-verbose* state)
  #+allegro (setf (sys:gsgc-switch switch) state))
